import torch
from srf_attention import simplex_random_matrix
from srf_attention import softmax_kernel_bhld, softmax_kernel_blhd
from functools import partial
try:
    import flash_attn_no_softmax
except:
    pass

def attn_no_softmax(q, k, v, causal):
    s = q @ k.transpose(-1, -2)
    if causal:
        s = s.tril()
    l = s.sum(-1)
    p = s / l[..., None]
    o = p @ v
    return o

class Attention(torch.nn.Module):
    def __init__(self, d, causal, n_features = None, scale = False, normalize = False, device = None):
        super().__init__()
        self.causal = causal
        self.n_features = d if n_features is None else n_features
        self.d = d
        self.scale = scale
        self.normalize = normalize
        self.redraw_on_call = False
        self.draw_features = partial(simplex_random_matrix, nb_rows = self.d, nb_columns = self.n_features, scale = self.scale, normalize = self.normalize)
        projection_matrix = self.draw_features(device = device)
        self.register_buffer('projection_matrix', projection_matrix)

    def redraw_on_call_(self, value=True):
        self.redraw_on_call = value

    def redraw_(self, device = None):
        projection = self.draw_features(device = device if device else self.projection_matrix.device)
        self.projection_matrix.copy_(projection)
        del projection

    # (B, H, L, D)
    def forward_train_torch(self, q, k, v):
        q = softmax_kernel_bhld(data=q, projection_matrix=self.projection_matrix, is_query=True, normalize_data=True, eps=1e-4)
        k, key_maxima = softmax_kernel_bhld(data=k, projection_matrix=self.projection_matrix, is_query=False, normalize_data=True, eps=1e-4)
        o = attn_no_softmax(q, k, v, causal=self.causal)
        return o

    # (B, L, H, D)
    def forward_train_flash(self, q, k, v):
        q = softmax_kernel_blhd(data=q, projection_matrix=self.projection_matrix, is_query=True, normalize_data=True, eps=1e-4)
        k, key_maxima = softmax_kernel_blhd(data=k, projection_matrix=self.projection_matrix, is_query=False, normalize_data=True, eps=1e-4)
        o = flash_attn_no_softmax.flash_attn_func(q, k, v, softmax_scale=1.0, causal=self.causal)
        return o

    # q, k, v should be of shape (b, h, l, d)
    def forward(self, q=None, k=None, v=None, mode = None, attn_fn = None):
        assert mode in ['train'] and attn_fn in ['torch', 'flash']
        if self.redraw_on_call:
            self.redraw_(device = q.device)
        
        if mode == 'train':
            if attn_fn == 'torch':
                o = self.forward_train_torch(q, k, v)
                return o
            elif attn_fn == 'flash':
                q = q.transpose(1, 2)
                k = k.transpose(1, 2)
                o = self.forward_train_flash(q, k, v)
                return o.transpose(1, 2)



import torch
from einops import repeat

def softmax_kernel_bhld(data, projection_matrix, is_query, normalize_data=True, eps=1e-4):
    b, h, l, d = data.shape

    data_normalizer = (data.shape[-1] ** -0.25) if normalize_data else 1.

    ratio = projection_matrix.shape[0] ** -0.5

    projection = repeat(projection_matrix, 'j d -> b h j d', b = b, h = h)
    projection = projection.type_as(data)

    data_dash = torch.einsum('...id,...jd->...ij', (data_normalizer * data), projection)

    diag_data = data.pow(2).sum(-1).mul(0.5 * data_normalizer ** 2).unsqueeze(-1)

    if is_query:
        data_dash = ratio * (
            torch.exp(data_dash - diag_data -
                    torch.amax(data_dash, dim=-1, keepdim=True).detach()) + eps)
        return data_dash.type_as(data)
    else:
        key_maxima = torch.amax(data_dash, dim=(-1, -2), keepdim=True).detach()
        data_dash = ratio * (
            torch.exp(data_dash - diag_data - key_maxima) + eps)
        return data_dash.type_as(data), key_maxima.type_as(data)

def softmax_kernel_blhd(data, projection_matrix, is_query, normalize_data=True, eps=1e-4):
    b, l, h, d = data.shape

    data_normalizer = (data.shape[-1] ** -0.25) if normalize_data else 1.

    ratio = projection_matrix.shape[0] ** -0.5

    projection = repeat(projection_matrix, 'j d -> b l j d', b = b, l = l)
    projection = projection.type_as(data)

    data_dash = torch.einsum('...id,...jd->...ij', (data_normalizer * data), projection)

    diag_data = data.pow(2).sum(-1).mul(0.5 * data_normalizer ** 2).unsqueeze(-1)

    if is_query:
        data_dash = ratio * (
            torch.exp(data_dash - diag_data -
                    torch.amax(data_dash, dim=-1, keepdim=True).detach()) + eps)
        return data_dash.type_as(data)
    else:
        key_maxima = torch.amax(data_dash, dim=(-1, -3), keepdim=True).detach()
        data_dash = ratio * (
            torch.exp(data_dash - diag_data - key_maxima) + eps)
        return data_dash.type_as(data), key_maxima.type_as(data)




from setuptools import setup, find_packages

setup(
    name='srf-attention',
    version='1.0.5',
    packages=find_packages(),
    author='Alex Levenston',
    author_email='alexlevenston2021@gmail.com',
    description='Simplex random feature attention in PyTorch for both training and inference',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/alexjlevenston/srf-attention',
    license='MIT',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    install_requires=[
        'torch',
        'scipy',
        'einops'
    ],
    include_package_data=True,
    python_requires='>=3.6',
)





import math
import torch
from scipy.linalg import hadamard

# generate a hadamard matrix of size d x d
@torch.no_grad()
def orthogonal_matrix_chunk_hadamard(d, device = None):
    (d1,d2,d3)=(torch.diag(torch.sign(torch.randn(d)-0.5)),
                torch.diag(torch.sign(torch.randn(d)-0.5)),
                torch.diag(torch.sign(torch.randn(d)-0.5)))
    (h1,h2,h3) =[torch.tensor(hadamard(d)).to(d1) for _ in range(3)]
    m = torch.matmul(torch.matmul(h3,d3),torch.matmul(torch.matmul(h2,d2),torch.matmul(h1,d1)))
    m = m / torch.sqrt(torch.sum(m[0]**2))
    m = m.to(device)
    return m

# compute dirs
@torch.no_grad()
def compute_simplex_dir(d, device = None):
    simp_dir = torch.diag(torch.ones(d))/math.sqrt(2) - torch.ones((d,d))*(1/((d-1) * math.sqrt(2))) *(1 + 1/math.sqrt(d)) #begin by getting the proj directions
    simp_dir[d-1,:] = 1/math.sqrt(2 * d) * torch.ones((d))
    simp_dir[:,d-1] = 0
    simp_dir = simp_dir / math.sqrt(torch.sum(simp_dir[1,:]**2))
    chi2 = torch.distributions.chi2.Chi2(torch.ones(d, dtype=torch.float32))
    rand_sim = torch.matmul(torch.diag(torch.sqrt(chi2.sample().to(simp_dir))),simp_dir)
    rand_sim = rand_sim.to(device)
    return rand_sim

# generate rfs
@torch.no_grad()
def simplex_random_matrix(nb_rows, nb_columns, scale = False, normalize = False, device = None):
    nb_full_blocks = int(nb_rows / nb_columns)

    block_list = []

    for _ in range(nb_full_blocks):
        q = orthogonal_matrix_chunk_hadamard(nb_columns, device = device)
        q = torch.matmul(compute_simplex_dir(nb_columns, device = device), q)
        block_list.append(q)

    remaining_rows = nb_rows - nb_full_blocks * nb_columns
    if remaining_rows > 0:
        q = orthogonal_matrix_chunk_hadamard(nb_columns, device = device)
        q = torch.matmul(compute_simplex_dir(nb_columns, device = device), q)
        block_list.append(q[:remaining_rows])

    final_matrix = torch.cat(block_list).to(device).float()
    if scale:
        final_matrix /= math.sqrt(nb_columns)
    if normalize:
        final_matrix /= final_matrix.norm(p=2, dim=-1)[..., None]
    return final_matrix

